@@ -1,40 +1,18 @@
 #!/bin/bash
+# SPDX-License-Identifier: GPL-2.0
+# Copyright (C) 2015-2020 Intel Corporation. All rights reserved.
 
-# Copyright(c) 2015-2017 Intel Corporation. All rights reserved.
-# 
-# This program is free software; you can redistribute it and/or modify it
-# under the terms of version 2 of the GNU General Public License as
-# published by the Free Software Foundation.
-# 
-# This program is distributed in the hope that it will be useful, but
-# WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-# General Public License for more details.
+. $(dirname $0)/common
 
-NDCTL="../ndctl/ndctl"
 rc=77
 
 set -e
 
-err() {
-	echo "test/device-dax-fio.sh: failed at line $1"
-	exit $rc
-}
+check_min_kver "4.11" || do_skip "kernel may lack device-dax fixes"
 
-check_min_kver()
-{
-	local ver="$1"
-	: "${KVER:=$(uname -r)}"
-
-	[ -n "$ver" ] || return 1
-	[[ "$ver" == "$(echo -e "$ver\n$KVER" | sort -V | head -1)" ]]
-}
-
-check_min_kver "4.11" || { echo "kernel $KVER may lack latest device-dax fixes"; exit $rc; }
-
-set -e
 trap 'err $LINENO' ERR
 
+check_prereq "fio"
 if ! fio --enghelp | grep -q "dev-dax"; then
 	echo "fio lacks dev-dax engine"
 	exit 77
@@ -79,7 +57,7 @@ do
 
 	# revert namespace to raw mode
 	json=$($NDCTL create-namespace -m raw -f -e $dev)
-	mode=$(echo $json | jq -r ".mode")
+	eval $(json2var <<< "$json")
 	[ $mode != "fsdax" ] && echo "fail: $LINENO" && exit 1
 done
 
